# Licensed under a 3-clause BSD style license, see LICENSE.

"""
Units in the HEP system of units
================================

This module `hepunits.units` contains 2 submodules for:

  * Basic physical units and derived units (`hepunits.units.units`).
  * Commonly-used unit prefixes (`hepunits.units.prefixes`).

All units are provided in the HEP System of Units as defined below.

In HEP the System of Units consists of the basic units originally defined by
the [CLHEP]_ project:

    ===================   ================== ====
    Quantity              Name               Unit
    ===================   ================== ====
    Length                millimeter         mm
    Time                  nanosecond         ns
    Energy                Mega electron Volt MeV
    Positron charge       eplus
    Temperature           kelvin             K
    Amount of substance   mole               mol
    Luminous intensity    candela            cd
    Plane angle           radian             rad
    Solid angle           steradian          sr
    ===================   ================== ====

It is largely based on the international system of units ([SI]_)

    ===================   ========   ====
    Quantity              Name       Unit
    ===================   ========   ====
    Length                meter      m
    Time                  second     s
    Mass                  kilogram   kg
    Electric current      ampere     A
    Temperature           kelvin     K
    Amount of substance   mole       mol
    Luminous intensity    candela    cd
    ===================   ========   ====

but augments it with handy definitions, changing the basic length and time
units.

Typical use cases::

    >>> # add two quantities with length units and get the result in meters
    >>> from hepunits import units as u
    >>> (1 * u.meter + 5 * u.cm) / u.meter
    1.05
    >>> from hepunits.units import MeV, GeV
    >>> massWindow = 100 * MeV    # define a 100 MeV mass window
    >>> def energy_resolution():
    ...    # returns the energy resolution of 100 MeV
    ...    return 100 * MeV
    ...
    >>> energy_resolution() / GeV # get the energy resolution in GeV
    0.1

**References**

.. [CLHEP] http://proj-clhep.web.cern.ch/proj-clhep/.
.. [SI] http://www.physics.nist.gov/cuu/Units/index.html.
"""

from math import pi as _pi
from typing import List

from . import prefixes as _pre

# --------------------------------------------------------------------
# Units of length
# ---------------
millimeter = 1.0
millimeter2 = millimeter * millimeter
millimeter3 = millimeter * millimeter * millimeter

mm = millimeter
mm2 = millimeter2
mm3 = millimeter3

meter = _pre.kilo * millimeter
meter2 = meter * meter
meter3 = meter * meter * meter

m = meter
m2 = meter2
m3 = meter3

centimeter = _pre.centi * meter
centimeter2 = centimeter * centimeter
centimeter3 = centimeter * centimeter * centimeter

cm = centimeter
cm2 = centimeter2
cm3 = centimeter3

kilometer = _pre.kilo * meter
kilometer2 = kilometer * kilometer
kilometer3 = kilometer * kilometer * kilometer

km = kilometer
km2 = kilometer2
km3 = kilometer3

micrometer = _pre.micro * meter
micron = micrometer

nanometer = _pre.nano * meter

angstrom = 1e-10 * meter

femtometer = _pre.femto * meter
fermi = femtometer

fm = femtometer
fm2 = femtometer * femtometer
fm3 = femtometer * femtometer * femtometer

barn = 1.0e-28 * meter2

millibarn = _pre.milli * barn
microbarn = _pre.micro * barn
nanobarn = _pre.nano * barn
picobarn = _pre.pico * barn
femtobarn = _pre.femto * barn
attobarn = _pre.atto * barn

mb = millibarn
ub = microbarn
nb = nanobarn
pb = picobarn
fb = femtobarn
ab = attobarn

invmb = 1.0 / millibarn
invub = 1.0 / microbarn
invnb = 1.0 / nanobarn
invpb = 1.0 / picobarn
invfb = 1.0 / femtobarn
invab = 1.0 / attobarn

# --------------------------------------------------------------------
# Units of time
# --------------------------------------------------------------------
nanosecond = 1.0

ns = nanosecond

second = _pre.giga * nanosecond
millisecond = _pre.milli * second
microsecond = _pre.micro * second
picosecond = _pre.pico * second
femtosecond = _pre.femto * second
attosecond = _pre.atto * second
zeptosecond = _pre.zepto * second
yoctosecond = _pre.yocto * second

s = second
ms = millisecond
us = microsecond
ps = picosecond
fs = femtosecond
# shortcut "as = attosecond" not possible since "as" is a reserved word ;-)
zs = zeptosecond
ys = yoctosecond

minute = 60 * second
hour = 60 * minute
day = 24 * hour
year = 365.25 * day

h = hour
d = day
y = year

hertz = 1.0 / second

kilohertz = _pre.kilo * hertz
megahertz = _pre.mega * hertz
gigahertz = _pre.giga * hertz
terahertz = _pre.tera * hertz

Hz = hertz

kHz = kilohertz
MHz = megahertz
GHz = gigahertz
THz = terahertz

# --------------------------------------------------------------------
# Units of energy
# --------------------------------------------------------------------
megaelectronvolt = 1.0

electronvolt = _pre.micro * megaelectronvolt

zettaelectronvolt = _pre.zetta * electronvolt
exaelectronvolt = _pre.exa * electronvolt
petaelectronvolt = _pre.peta * electronvolt
teraelectronvolt = _pre.tera * electronvolt
gigaelectronvolt = _pre.giga * electronvolt
kiloelectronvolt = _pre.kilo * electronvolt

ZeV = zettaelectronvolt
EeV = exaelectronvolt
PeV = petaelectronvolt
TeV = teraelectronvolt
GeV = gigaelectronvolt
MeV = megaelectronvolt
keV = kiloelectronvolt
eV = electronvolt

# --------------------------------------------------------------------
# Units of electric charge
# --------------------------------------------------------------------
eplus = 1.0  # positron charge

# --------------------------------------------------------------------
# Units of temperature
# --------------------------------------------------------------------
kelvin = 1.0

K = kelvin

# --------------------------------------------------------------------
# Units of amount of substance
# --------------------------------------------------------------------
mole = 1.0

mol = mole

# --------------------------------------------------------------------
# Units of luminous intensity
# --------------------------------------------------------------------
candela = 1.0

cd = candela

# --------------------------------------------------------------------
# Units of angles
# --------------------------------------------------------------------
radian = 1.0  # plane angle
steradian = 1.0  # solid angle

rad = radian
sr = steradian

milliradian = _pre.milli * radian
mrad = milliradian

degree = (_pi / 180.0) * radian

deg = degree

# --------------------------------------------------------------------
# Derived units
# --------------------------------------------------------------------

# Positron charge [Coulomb]
e_SI = 1.602176634e-19  # exact value, taken from PDG 2022


# Electric charge [Q]
# --------------------------------------------------------------------
coulomb = eplus / e_SI

# Electric current [Q][T^-1]
# -------------------------
ampere = coulomb / second

milliampere = _pre.milli * ampere
microampere = _pre.micro * ampere
nanoampere = _pre.nano * ampere

A = ampere

# Energy [E]
# ----------
joule = electronvolt / e_SI  # joule = 6.24151e+12 * MeV

J = joule

erg = 1.0e-7 * joule

# Power [M][L^2][T^-3]
watt = joule / second

W = watt

kW = _pre.kilo * watt
MW = _pre.mega * watt
GW = _pre.giga * watt

# Force [E][L^-1]
newton = joule / meter

N = newton

dyne = 1.0e-5 * newton

# Pressure
pascal = newton / meter2

Pa = pascal

bar = 1.0e5 * pascal

atmosphere = 101325 * pascal

# Mass [E][T^2][L^-2]
kilogram = joule * second * second / (meter * meter)
gram = _pre.milli * kilogram
milligram = _pre.milli * gram

kg = kilogram
g = gram
mg = milligram

# Electric potential
megavolt = megaelectronvolt / eplus
volt = _pre.micro * megavolt
kilovolt = _pre.kilo * volt

# Electric capacitance
farad = coulomb / volt

millifarad = _pre.milli * farad
microfarad = _pre.micro * farad
nanofarad = _pre.nano * farad
picofarad = _pre.pico * farad

# Electric resistance
ohm = volt / ampere

# Magnetic Field
tesla = volt * second / meter2

gauss = 1.0e-4 * tesla

kilogauss = _pre.kilo * gauss

# Magnetic Flux
weber = volt * second  # weber = 1000*megavolt*ns

milliweber = _pre.milli * weber
microweber = _pre.micro * weber
nanoweber = _pre.nano * weber

Wb = weber

mWb = milliweber
uWb = microweber
nWb = nanoweber

# Inductance
henry = weber / ampere

H = henry

# --------------------------------------------------------------------
# Units derived from luminous intensity
# --------------------------------------------------------------------

# Luminous flux [I]
lumen = candela * steradian

# Illuminance, i.e. amount of luminous flux per unit area [I][L^-2]
lux = lumen / meter2

# --------------------------------------------------------------------
# Units for radiation
# --------------------------------------------------------------------

# Activity [T^-1]
becquerel = 1.0 / second

kilobecquerel = _pre.kilo * becquerel
megabecquerel = _pre.mega * becquerel
gigabecquerel = _pre.giga * becquerel

Bq = becquerel

kBq = kilobecquerel
MBq = megabecquerel
GBq = gigabecquerel

curie = 3.7e10 * becquerel

millicurie = _pre.milli * curie
microcurie = _pre.micro * curie
nanocurie = _pre.nano * curie

Ci = curie

mCi = millicurie
uCi = microcurie
nCi = nanocurie

# Absorbed dose [L^2][T^-2]
gray = joule / kilogram

megagray = _pre.mega * gray
kilogray = _pre.kilo * gray
milligray = _pre.milli * gray
microgray = _pre.micro * gray

Gy = gray

MGy = megagray
kGy = kilogray
mGy = milligray
uGy = microgray

# Dose equivalent
sievert = joule / kilogram

Sv = sievert

__all__ = (
    "GW",
    "MW",
    "A",
    "Bq",
    "Ci",
    "EeV",
    "GBq",
    "GHz",
    "GeV",
    "Gy",
    "H",
    "Hz",
    "J",
    "K",
    "MBq",
    "MGy",
    "MHz",
    "MeV",
    "N",
    "Pa",
    "PeV",
    "Sv",
    "THz",
    "TeV",
    "W",
    "Wb",
    "ZeV",
    "ab",
    "ampere",
    "angstrom",
    "atmosphere",
    "attobarn",
    "attosecond",
    "bar",
    "barn",
    "becquerel",
    "candela",
    "cd",
    "centimeter",
    "centimeter2",
    "centimeter3",
    "cm",
    "cm2",
    "cm3",
    "coulomb",
    "curie",
    "d",
    "day",
    "deg",
    "degree",
    "dyne",
    "eV",
    "e_SI",
    "electronvolt",
    "eplus",
    "erg",
    "exaelectronvolt",
    "farad",
    "fb",
    "femtobarn",
    "femtometer",
    "femtosecond",
    "fermi",
    "fm",
    "fm2",
    "fm3",
    "fs",
    "g",
    "gauss",
    "gigabecquerel",
    "gigaelectronvolt",
    "gigahertz",
    "gram",
    "gray",
    "h",
    "henry",
    "hertz",
    "hour",
    "invab",
    "invfb",
    "invmb",
    "invnb",
    "invpb",
    "invub",
    "joule",
    "kBq",
    "kGy",
    "kHz",
    "kW",
    "keV",
    "kelvin",
    "kg",
    "kilobecquerel",
    "kiloelectronvolt",
    "kilogauss",
    "kilogram",
    "kilogray",
    "kilohertz",
    "kilometer",
    "kilometer2",
    "kilometer3",
    "kilovolt",
    "km",
    "km2",
    "km3",
    "lumen",
    "lux",
    "m",
    "m2",
    "m3",
    "mCi",
    "mGy",
    "mWb",
    "mb",
    "megabecquerel",
    "megaelectronvolt",
    "megagray",
    "megahertz",
    "megavolt",
    "meter",
    "meter2",
    "meter3",
    "mg",
    "microampere",
    "microbarn",
    "microcurie",
    "microfarad",
    "microgray",
    "micrometer",
    "micron",
    "microsecond",
    "microweber",
    "milliampere",
    "millibarn",
    "millicurie",
    "millifarad",
    "milligram",
    "milligray",
    "millimeter",
    "millimeter2",
    "millimeter3",
    "milliradian",
    "millisecond",
    "milliweber",
    "minute",
    "mm",
    "mm2",
    "mm3",
    "mol",
    "mole",
    "mrad",
    "ms",
    "nCi",
    "nWb",
    "nanoampere",
    "nanobarn",
    "nanocurie",
    "nanofarad",
    "nanometer",
    "nanosecond",
    "nanoweber",
    "nb",
    "newton",
    "ns",
    "ohm",
    "pascal",
    "pb",
    "petaelectronvolt",
    "picobarn",
    "picofarad",
    "picosecond",
    "ps",
    "rad",
    "radian",
    "s",
    "second",
    "sievert",
    "sr",
    "steradian",
    "teraelectronvolt",
    "terahertz",
    "tesla",
    "uCi",
    "uGy",
    "uWb",
    "ub",
    "us",
    "volt",
    "watt",
    "weber",
    "y",
    "year",
    "yoctosecond",
    "ys",
    "zeptosecond",
    "zettaelectronvolt",
    "zs",
)


def __dir__() -> List[str]:
    return list(__all__)



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Physical and other handy constants
==================================

This module `hepunits.constants` contains 2 sorts of constants:

  * Physical constants.
  * Common and/or handy constants.

All constants are computed in the HEP System of Units
as defined in the `hepunits.units` module.

Typical use case::

    >>> from hepunits.constants import c_light
    >>> from hepunits.units     import picosecond, micrometer
    >>> tau_Bs = 1.5 * picosecond    # a particle lifetime, say the Bs meson's
    >>> ctau_Bs = c_light * tau_Bs   # ctau of the particle, ~450 microns
    >>> print ctau_Bs                # result in HEP units, so mm ;-)
    0.449688687
    >>> print ctau_Bs / micrometer   # result in micrometers
    449.688687

"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------

from math import pi
from typing import List

from ..units.units import e_SI, electronvolt, eplus, joule, kelvin, m, mole, s

__all__ = (
    "Avogadro",
    "c_light",
    "c_light_sq",
    "e_SI",
    "e_sq",
    "electronvolt",
    "eminus",
    "eplus",
    "h_Planck",
    "half_pi",
    "hbar",
    "hbar_Planck",
    "hbarc",
    "hbarc_sq",
    "joule",
    "k_Boltzmann",
    "kelvin",
    "m",
    "mole",
    "pi",
    "pi_sq",
    "s",
    "two_pi",
)


# -----------------------------------------------------------------------------
# Mathematical constants
# -----------------------------------------------------------------------------

two_pi = 2 * pi
half_pi = pi / 2
pi_sq = pi * pi

# -----------------------------------------------------------------------------
# Physical constants
# -----------------------------------------------------------------------------

# Speed of light in vacuum (exact value, taken from PDG 2022)
c_light = 299792458 * m / s
c_light_sq = c_light * c_light

# Electron charge
eminus = -eplus  # electron charge
e_sq = eplus * eplus

# Avogadro constant (exact value, taken from PDG 2022)
Avogadro = 6.02214076e23 / mole

# Planck constant (exact value, taken from PDG 2022)
# h_Planck = 6.62607015e-34 * joule * s
# Simplify the units to avoid floating errors, using joule = electronvolt / e_SI,
# noticing also that electronvolt * s = 1000.0 to rearrange
h_Planck = 6.62607015e-34 * electronvolt * s / e_SI

hbar_Planck = h_Planck / two_pi
hbar = hbar_Planck

hbarc = hbar_Planck * c_light

hbarc_sq = hbarc * hbarc

# Boltzmann constant (exact value, taken from PDG 2022)
k_Boltzmann = 1.380649e-23 * joule / kelvin


def __dir__() -> List[str]:
    return list(__all__)



# Licensed under a 3-clause BSD style license, see LICENSE.

from typing import List

from . import prefixes, units
from .prefixes import (
    atto,
    centi,
    deca,
    deci,
    exa,
    exbi,
    femto,
    gibi,
    giga,
    googol,
    hecto,
    kibi,
    kilo,
    mebi,
    mega,
    micro,
    milli,
    nano,
    pebi,
    peta,
    pico,
    tebi,
    tera,
    yobi,
    yocto,
    yotta,
    zebi,
    zepto,
    zetta,
)
from .units import (
    GW,
    MW,
    A,
    Bq,
    Ci,
    EeV,
    GBq,
    GeV,
    GHz,
    Gy,
    H,
    Hz,
    J,
    K,
    MBq,
    MeV,
    MGy,
    MHz,
    N,
    Pa,
    PeV,
    Sv,
    TeV,
    THz,
    W,
    Wb,
    ZeV,
    ab,
    ampere,
    angstrom,
    atmosphere,
    attobarn,
    attosecond,
    bar,
    barn,
    becquerel,
    candela,
    cd,
    centimeter,
    centimeter2,
    centimeter3,
    cm,
    cm2,
    cm3,
    coulomb,
    curie,
    d,
    day,
    deg,
    degree,
    dyne,
    e_SI,
    electronvolt,
    eplus,
    erg,
    eV,
    exaelectronvolt,
    farad,
    fb,
    femtobarn,
    femtometer,
    femtosecond,
    fermi,
    fm,
    fm2,
    fm3,
    fs,
    g,
    gauss,
    gigabecquerel,
    gigaelectronvolt,
    gigahertz,
    gram,
    gray,
    h,
    henry,
    hertz,
    hour,
    invab,
    invfb,
    invmb,
    invnb,
    invpb,
    invub,
    joule,
    kBq,
    kelvin,
    keV,
    kg,
    kGy,
    kHz,
    kilobecquerel,
    kiloelectronvolt,
    kilogauss,
    kilogram,
    kilogray,
    kilohertz,
    kilometer,
    kilometer2,
    kilometer3,
    kilovolt,
    km,
    km2,
    km3,
    kW,
    lumen,
    lux,
    m,
    m2,
    m3,
    mb,
    mCi,
    megabecquerel,
    megaelectronvolt,
    megagray,
    megahertz,
    megavolt,
    meter,
    meter2,
    meter3,
    mg,
    mGy,
    microampere,
    microbarn,
    microcurie,
    microfarad,
    microgray,
    micrometer,
    micron,
    microsecond,
    microweber,
    milliampere,
    millibarn,
    millicurie,
    millifarad,
    milligram,
    milligray,
    millimeter,
    millimeter2,
    millimeter3,
    milliradian,
    millisecond,
    milliweber,
    minute,
    mm,
    mm2,
    mm3,
    mol,
    mole,
    mrad,
    ms,
    mWb,
    nanoampere,
    nanobarn,
    nanocurie,
    nanofarad,
    nanometer,
    nanosecond,
    nanoweber,
    nb,
    nCi,
    newton,
    ns,
    nWb,
    ohm,
    pascal,
    pb,
    petaelectronvolt,
    picobarn,
    picofarad,
    picosecond,
    ps,
    rad,
    radian,
    s,
    second,
    sievert,
    sr,
    steradian,
    teraelectronvolt,
    terahertz,
    tesla,
    ub,
    uCi,
    uGy,
    us,
    uWb,
    volt,
    watt,
    weber,
    y,
    year,
    yoctosecond,
    ys,
    zeptosecond,
    zettaelectronvolt,
    zs,
)

__all__ = (
    "GW",
    "MW",
    "A",
    "Bq",
    "Ci",
    "EeV",
    "GBq",
    "GHz",
    "GeV",
    "Gy",
    "H",
    "Hz",
    "J",
    "K",
    "MBq",
    "MGy",
    "MHz",
    "MeV",
    "N",
    "Pa",
    "PeV",
    "Sv",
    "THz",
    "TeV",
    "W",
    "Wb",
    "ZeV",
    "ab",
    "ampere",
    "angstrom",
    "atmosphere",
    "atto",
    "attobarn",
    "attosecond",
    "bar",
    "barn",
    "becquerel",
    "candela",
    "cd",
    "centi",
    "centimeter",
    "centimeter2",
    "centimeter3",
    "cm",
    "cm2",
    "cm3",
    "coulomb",
    "curie",
    "d",
    "day",
    "deca",
    "deci",
    "deg",
    "degree",
    "dyne",
    "eV",
    "e_SI",
    "electronvolt",
    "eplus",
    "erg",
    "exa",
    "exaelectronvolt",
    "exbi",
    "farad",
    "fb",
    "femto",
    "femtobarn",
    "femtometer",
    "femtosecond",
    "fermi",
    "fm",
    "fm2",
    "fm3",
    "fs",
    "g",
    "gauss",
    "gibi",
    "giga",
    "gigabecquerel",
    "gigaelectronvolt",
    "gigahertz",
    "googol",
    "gram",
    "gray",
    "h",
    "hecto",
    "henry",
    "hertz",
    "hour",
    "invab",
    "invfb",
    "invmb",
    "invnb",
    "invpb",
    "invub",
    "joule",
    "kBq",
    "kGy",
    "kHz",
    "kW",
    "keV",
    "kelvin",
    "kg",
    "kibi",
    "kilo",
    "kilobecquerel",
    "kiloelectronvolt",
    "kilogauss",
    "kilogram",
    "kilogray",
    "kilohertz",
    "kilometer",
    "kilometer2",
    "kilometer3",
    "kilovolt",
    "km",
    "km2",
    "km3",
    "lumen",
    "lux",
    "m",
    "m2",
    "m3",
    "mCi",
    "mGy",
    "mWb",
    "mb",
    "mebi",
    "mega",
    "megabecquerel",
    "megaelectronvolt",
    "megagray",
    "megahertz",
    "megavolt",
    "meter",
    "meter2",
    "meter3",
    "mg",
    "micro",
    "microampere",
    "microbarn",
    "microcurie",
    "microfarad",
    "microgray",
    "micrometer",
    "micron",
    "microsecond",
    "microweber",
    "milli",
    "milliampere",
    "millibarn",
    "millicurie",
    "millifarad",
    "milligram",
    "milligray",
    "millimeter",
    "millimeter2",
    "millimeter3",
    "milliradian",
    "millisecond",
    "milliweber",
    "minute",
    "mm",
    "mm2",
    "mm3",
    "mol",
    "mole",
    "mrad",
    "ms",
    "nCi",
    "nWb",
    "nano",
    "nanoampere",
    "nanobarn",
    "nanocurie",
    "nanofarad",
    "nanometer",
    "nanosecond",
    "nanoweber",
    "nb",
    "newton",
    "ns",
    "ohm",
    "pascal",
    "pb",
    "pebi",
    "peta",
    "petaelectronvolt",
    "pico",
    "picobarn",
    "picofarad",
    "picosecond",
    "prefixes",
    "ps",
    "rad",
    "radian",
    "s",
    "second",
    "sievert",
    "sr",
    "steradian",
    "tebi",
    "tera",
    "teraelectronvolt",
    "terahertz",
    "tesla",
    "uCi",
    "uGy",
    "uWb",
    "ub",
    "units",
    "us",
    "volt",
    "watt",
    "weber",
    "y",
    "year",
    "yobi",
    "yocto",
    "yoctosecond",
    "yotta",
    "ys",
    "zebi",
    "zepto",
    "zeptosecond",
    "zetta",
    "zettaelectronvolt",
    "zs",
)


def __dir__() -> List[str]:
    return list(__all__)



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Conversion routines between CLHEP, the HEP System of Units, and Pint units.

CLHEP adopts the approach where all the quantities are stored in the base unit
system, effectively dimensionless. Pint offers to store both the magnitude and
the dimensionality of the unit, which is helpful in deducing and/or validating
the resulting unit of formulas. This module offers conversion routines between
Pint's default base unit system and CLHEP.
"""

from __future__ import annotations

try:
    import pint
except ImportError as exc:
    msg = "Pint is required to use hepunits.pint."
    raise ImportError(msg) from exc

# TODO: support more unit conversions
_clhep_base_units = {
    "[length]": "millimeter",
    "[time]": "nanosecond",
    "[mass]": "MeV * millimeter**-2 * nanosecond**2",
    "[current]": "elementary_charge / nanosecond",
}


def _unit_from(val: pint.Quantity | pint.Unit) -> pint.Unit:
    """Extract the dimensionality from a Pint Quantity or Unit."""
    # Grabbing the type is a quick way to be in the correct unit registry
    # see e.g. https://github.com/hgrecco/pint/issues/2207
    unit = type(val.units) if isinstance(val, pint.Quantity) else type(val)
    out = unit("dimensionless")
    for dim, exponent in val.dimensionality.items():
        if dim not in _clhep_base_units:
            msg = f"Unsupported dimension {dim} in {val}"
            raise ValueError(msg)
        out *= unit(_clhep_base_units[dim]) ** exponent

    return out


def to_clhep(val: pint.Quantity | pint.Unit) -> float:
    """
    Convert a Pint Quantity or Unit to CLHEP base units.

    Parameters
    ----------
    val : pint.Quantity or pint.Unit
        The value to convert.

    Returns
    -------
    float
        The value in CLHEP base units (dimensionless).

    Examples
    --------
    >>> ureg = pint.UnitRegistry()
    >>> g = 9.8 * ureg.meter / ureg.second**2
    >>> g
    <Quantity(9.8, 'meter / second ** 2')>
    >>> to_clhep(g)
    9.800000000000001e-15
    """
    clhep_unit = _unit_from(val)
    q = pint.Quantity(1.0, val) if isinstance(val, pint.Unit) else val
    return q.to(clhep_unit).magnitude  # type: ignore[no-any-return]


def from_clhep(val: float, unit: pint.Unit) -> pint.Quantity:
    """
    Convert a value in CLHEP base units to a Pint Quantity.

    Parameters
    ----------
    val : float
        The value in CLHEP base units (dimensionless).
    unit : pint.Unit
        The desired output unit.

    Returns
    -------
    pint.Quantity
        The value in the desired unit.

    Examples
    --------
    >>> ureg = pint.UnitRegistry()
    >>> from_clhep(hepunits.c_light, ureg.meter / ureg.second)
    <Quantity(299792458.0, 'meter / second')>
    """
    clhep_unit = _unit_from(unit)
    return pint.Quantity(val, clhep_unit).to(unit)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the hepunits.units.prefixes module.
"""

from math import log

from pytest import approx

from hepunits import kibi, mega, micro, tebi, yocto, yotta


def test_prefixes_e6():
    assert 4 * mega == 1.0 / 0.25 / micro


def test_prefixes_e24():
    assert yotta * yocto == approx(1.0)


def test_prefixes_binary():
    assert log(kibi, 2) == 10
    assert log(tebi, 2) == 40



import nox

nox.needs_version = ">=2024.4.15"
nox.options.default_venv_backend = "uv|virtualenv"


@nox.session
def lint(session: nox.Session) -> None:
    """
    Run the linter.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """
    Run the unit and regular tests.
    """
    session.install("-e.[test]")
    session.run("pytest", *session.posargs)
